"""
bengrn base module.
"""

import pandas as pd
import urllib.request
import os.path

from grnndata import GRNAnnData, from_adata_and_longform, from_scope_loomfile, utils
from anndata import AnnData, concat
from typing import Optional
from .tools import GENIE3
import numpy as np

from arboreto.algo import grnboost2

from pyscenic.utils import modules_from_adjacencies
from pyscenic.prune import prune2df, df2regulons
from pyscenic.aucell import aucell

from ctxcore.rnkdb import FeatherRankingDatabase as RankingDatabase

import seaborn as sns
import decoupler as dc
import scipy.stats
from sklearn.metrics import precision_recall_curve, auc, PrecisionRecallDisplay


# example constant variable
NAME = "bengrn"

FILEDIR = os.path.dirname(os.path.realpath(__file__))


class BenGRN:
    def __init__(self, grn: GRNAnnData, full_dataset: Optional[AnnData] = None):
        """
        Initializes the BenGRN class.

        Parameters:
            grn (GRNAnnData): The Gene Regulatory Network data.
            full_dataset (Optional[AnnData]): The full dataset, defaults to None.
        """
        self.grn = grn
        self.full_dataset = full_dataset

    def do_tests(self, to: str = "collectri", organism="human"):
        """
        This method performs tests on the Gene Regulatory Network (GRN) and the full dataset.
        It compares the GRN network structure with the ground truth (GT) database and computes various metrics.

        Parameters:
            to (str, optional): The name of the GT database to use. Defaults to "collectri".
            organism (str, optional): The organism to consider for the GT database. Defaults to "human".

        Returns:
            dict: A dictionary containing the computed metrics.
        """
        self.gt = get_GT_db(name=to, organism=organism)
        # similarity in network structure
        # Get the GRN network from both grn objects
        grn_self = self.grn.grn
        grn_gt = self.gt.grn

        intersection = grn_self.index.intersection(grn_gt.index)
        print("intersection of {} genes".format(intersection.size))
        res = {"intersection": intersection.size / grn_self.index.size}
        intersection = list(intersection)
        grn_self = grn_self.loc[intersection, intersection]
        grn_gt = grn_gt.loc[intersection, intersection]
        similar_edges = (
            (grn_self != 0) & (grn_gt != 0) & (np.sign(grn_self) == np.sign(grn_gt))
        )
        similar_edges_ct = np.sum(np.sum(similar_edges))

        # Compute the total number of edges
        total_edges = np.sum(np.sum(grn_self != 0))

        # Compute the total number of edges in the other GRN
        total_edges_other = np.sum(np.sum((grn_gt != 0)))
        precision = similar_edges_ct / total_edges
        recall = similar_edges_ct / total_edges_other
        accuracy = similar_edges_ct / (
            total_edges + total_edges_other - similar_edges_ct
        )

        # Compute the Spearman's rank correlation between the two overlapping sets of edges
        spearman_corr = scipy.stats.spearmanr(
            grn_self.values[similar_edges].flatten(),
            grn_gt.values[similar_edges].flatten(),
        )
        res.update(
            {
                "precision": precision,
                "recall": recall,
                "accuracy": accuracy,
                "spearman_corr": spearman_corr,
            }
        )

        gt = grn_gt.values.flatten() != 0
        preval = gt.sum() / len(gt)
        pr, re, _ = precision_recall_curve(gt, grn_self.values.flatten())
        res.update({"auc": auc(re, pr)})
        disp = PrecisionRecallDisplay(pr, re, prevalence_pos_label=preval)
        disp.plot(plot_chance_level=True)
        disp.ax_.set_ylim([0, min(pr[:-1].max() * 3, 1)])

        return res

    def get_self_metrics(self):
        print("I'm getting my own metrics")
        pass

        # recap N protein complexes

        # gene pairs reg correlation with high string score vs none

        # recap well-known relationships

        #

    def compare_to(self, other: GRNAnnData):
        print("I'm comparing myself to another dataset")
        pass


def get_scenicplus(
    filepath=FILEDIR + "/../data/10xPBMC_homo_scenicplus_genebased_scope.loom",
):
    """
    This function retrieves a loomx scenicplus data from a given file path and loads it as a GrnnData

    Parameters:
        filepath : str, optional
            The path to the scenicplus data file.
            Default is FILEDIR + "/../data/10xPBMC_homo_scenicplus_genebased_scope.loom".

    Raises:
        FileNotFoundError: If the file at the given path does not exist.

    Returns:
        GrnAnnData: The scenicplus data from the given file as a grnndata object
    """
    if not os.path.exists(filepath):
        raise FileNotFoundError(
            f"The file {filepath} does not exist. You likely need to download \
                this or another loomxfile from the scope website"
        )

    return from_scope_loomfile(filepath)


def get_sroy_gt(join="outer"):
    """
    This function retrieves the ground truth data from Stone and Sroy's study.

    Parameters:
        join : str, optional
            The type of join to be performed when concatenating the data.
            Default is "outer".

    Returns:
        GrnAnnData: The ground truth data as a grnndata object
    """
    adata_liu = AnnData(
        (
            2
            ** pd.read_csv(
                FILEDIR
                + "/../data/GroundTruth/stone_and_sroy/scRNA/liu_rna_filtered_log2.tsv",
                sep="\t",
            )
        )
        - 1
    ).T
    adata_chen = AnnData(
        (
            2
            ** pd.read_csv(
                FILEDIR
                + "../data/GroundTruth/stone_and_sroy/scRNA/chen_rna_filtered_log2.tsv",
                sep="\t",
            )
        )
        - 1
    ).T
    df = pd.read_csv(
        FILEDIR + "/../data/GroundTruth/stone_and_sroy/hESC_ground_truth.tsv",
        sep="\t",
        header=None,
    )
    adata_liu.obs["dataset"] = "liu"
    adata_chen.obs["dataset"] = "chen"
    adata = concat([adata_liu, adata_chen], join=join)
    return from_adata_and_longform(adata, df)


def compute_scenic(adata, data_dir=FILEDIR + "/../data"):
    """
    This function computes the SCENIC algorithm on the given data.

    Parameters:
        adata (AnnData): The annotated data matrix of shape n_obs x n_vars. Rows correspond to cells and columns to genes.
        data_dir (str, optional): The directory where the data files will be stored. Defaults to FILEDIR + "/../data".

    Returns:
        GRNAnnData: The Gene Regulatory Network data.
    """
    os.makedirs(data_dir, exist_ok=True)

    url1 = "https://resources.aertslab.org/cistarget/motif2tf/motifs-v10nr_clust-nr.hgnc-m0.001-o0.0.tbl"
    url2 = "https://resources.aertslab.org/cistarget/databases/homo_sapiens/hg38/refseq_r80/mc_v10_clust/gene_based/hg38_10kbp_up_10kbp_down_full_tx_v10_clust.genes_vs_motifs.rankings.feather"

    file1 = os.path.join(data_dir, os.path.basename(url1))
    file2 = os.path.join(data_dir, os.path.basename(url2))

    if not os.path.exists(file1):
        urllib.request.urlretrieve(url1, file1)
    if not os.path.exists(file2):
        urllib.request.urlretrieve(url2, file2)
    dbs = RankingDatabase(fname=file2, name="genes_10kb_human_rank_v10")

    adjacencies = grnboost2(
        expression_data=adata.to_df(), tf_names=utils.TF, verbose=True
    )
    modules = list(modules_from_adjacencies(adjacencies, adata.to_df()))

    df = prune2df([dbs], modules, file1)
    regulons = df2regulons(df)
    # auc_mtx = aucell(adata.to_df(), regulons, num_workers=8)
    # sns.clustermap(auc_mtx, figsize=(12,12))

    # compute the grn matrix
    var_names = adata.var_names.tolist()
    da = np.zeros((len(var_names), len(var_names)), dtype=np.float)
    for reg in regulons:
        i = var_names.index(reg.transcription_factor)
        for k, v in reg.gene2weight.items():
            da[i, var_names.index(k)] = v
    # create the grndata
    grn = GRNAnnData(adata, grn=da)

    da = np.zeros((len(var_names), len(var_names)), dtype=np.float)
    for reg in modules:
        i = var_names.index(reg.transcription_factor)
        for k, v in reg.gene2weight.items():
            da[i, var_names.index(k)] = v
    grn.varp["modules"] = da
    return grn


def compute_genie3(adata, nthreads=30, ntrees=100, **kwargs):
    """
    This function computes the GENIE3 algorithm on the given data.

    Parameters:
        adata (AnnData): The annotated data matrix of shape n_obs x n_vars. Rows correspond to cells and columns to genes.
        nthreads (int, optional): The number of threads to use for computation. Defaults to 30.
        ntrees (int, optional): The number of trees to use for the Random Forests. Defaults to 100.
        **kwargs: Additional arguments to pass to the GENIE3 function.

    Returns:
        GRNAnnData: The Gene Regulatory Network data computed using the GENIE3 algorithm.
    """
    mat = np.asarray(adata.X.todense())
    names = adata.var_names[mat.sum(0) > 0].tolist()
    var = adata.var[mat.sum(0) > 0]
    mat = mat[:, mat.sum(0) > 0]
    VIM = GENIE3(mat, gene_names=names, nthreads=nthreads, ntrees=ntrees, **kwargs)
    return GRNAnnData(grn=VIM, X=mat, var=var, obs=adata.obs)


def get_GT_db(name="collectri", cell_type=None, organism="human", split_complexes=True):
    """
    use_prior_network loads a prior GRN from a list of available networks.

    Args:
        name (str, optional): name of the network to load. Defaults to "collectri".
        organism (str, optional): organism to load the network for. Defaults to "human".
        split_complexes (bool, optional): whether to split complexes into individual genes. Defaults to True.

    Raises:
        ValueError: if the provided name is not amongst the available names.
    """
    # TODO: use omnipath instead
    if name == "tflink":
        TFLINK = "https://cdn.netbiol.org/tflink/download_files/TFLink_Homo_sapiens_interactions_All_simpleFormat_v1.0.tsv.gz"
        net = pd_load_cached(TFLINK)
        net = net.rename(columns={"Name.TF": "source", "Name.Target": "target"})
    elif name == "htftarget":
        HTFTARGET = "http://bioinfo.life.hust.edu.cn/static/hTFtarget/file_download/tf-target-infomation.txt"
        net = pd_load_cached(HTFTARGET)
        net = net.rename(columns={"TF": "source"})
    elif name == "collectri":
        net = dc.get_collectri(organism=organism, split_complexes=split_complexes).drop(
            columns=["PMID"]
        )
    elif name == "dorothea":
        net = dc.get_dorothea(organism=organism)
    elif name == "regulonDB":
        pass
    else:
        raise ValueError(f"provided name: '{name}' is not amongst the available names.")
    varnames = list(set(net.iloc[:, :2].values.flatten()))
    adata = AnnData(var=varnames)
    adata.var_names = varnames
    grn = from_adata_and_longform(adata, net, has_weight=True)
    return grn


def pd_load_cached(url, loc="/tmp/", cache=True, **kwargs):
    # Check if the file exists, if not, download it
    loc += url.split("/")[-1]
    if not os.path.isfile(loc) or not cache:
        urllib.request.urlretrieve(url, loc)
    # Load the data from the file
    return pd.read_csv(loc, **kwargs)



from sklearn.tree import BaseDecisionTree
from sklearn.ensemble import RandomForestRegressor, ExtraTreesRegressor
from numpy import *
import time
from operator import itemgetter
from multiprocessing import Pool


def compute_feature_importances(estimator):
    if isinstance(estimator, BaseDecisionTree):
        return estimator.tree_.compute_feature_importances(normalize=False)
    else:
        importances = [
            e.tree_.compute_feature_importances(normalize=False)
            for e in estimator.estimators_
        ]
        importances = array(importances)
        return sum(importances, axis=0) / len(estimator)


def get_link_list(
    VIM, gene_names=None, regulators="all", maxcount="all", file_name=None
):
    """Gets the ranked list of (directed) regulatory links.

    Parameters
    ----------

    VIM: numpy array
        Array as returned by the function GENIE3(), in which the element (i,j) is the score of the edge directed from the i-th gene to the j-th gene.

    gene_names: list of strings, optional
        List of length p, where p is the number of rows/columns in VIM, containing the names of the genes. The i-th item of gene_names must correspond to the i-th row/column of VIM. When the gene names are not provided, the i-th gene is named Gi.
        default: None

    regulators: list of strings, optional
        List containing the names of the candidate regulators. When a list of regulators is provided, the names of all the genes must be provided (in gene_names), and the returned list contains only edges directed from the candidate regulators. When regulators is set to 'all', any gene can be a candidate regulator.
        default: 'all'

    maxcount: 'all' or positive integer, optional
        Writes only the first maxcount regulatory links of the ranked list. When maxcount is set to 'all', all the regulatory links are written.
        default: 'all'

    file_name: string, optional
        Writes the ranked list of regulatory links to the file file_name.
        default: None



    Returns
    -------

    The list of regulatory links, ordered according to the edge score. Auto-regulations do not appear in the list. Regulatory links with a score equal to zero are randomly permuted. In the ranked list of edges, each line has format:

        regulator   target gene     score of edge
    """

    # Check input arguments
    if not isinstance(VIM, ndarray):
        raise ValueError("VIM must be a square array")
    elif VIM.shape[0] != VIM.shape[1]:
        raise ValueError("VIM must be a square array")

    ngenes = VIM.shape[0]

    if gene_names is not None:
        if not isinstance(gene_names, (list, tuple)):
            raise ValueError("input argument gene_names must be a list of gene names")
        elif len(gene_names) != ngenes:
            raise ValueError(
                "input argument gene_names must be a list of length p, where p is the number of columns/genes in the expression data"
            )

    if regulators != "all":
        if not isinstance(regulators, (list, tuple)):
            raise ValueError("input argument regulators must be a list of gene names")

        if gene_names is None:
            raise ValueError(
                "the gene names must be specified (in input argument gene_names)"
            )
        else:
            sIntersection = set(gene_names).intersection(set(regulators))
            if not sIntersection:
                raise ValueError(
                    "The genes must contain at least one candidate regulator"
                )

    if maxcount != "all" and not isinstance(maxcount, int):
        raise ValueError('input argument maxcount must be "all" or a positive integer')

    if file_name is not None and not isinstance(file_name, str):
        raise ValueError("input argument file_name must be a string")

    # Get the indices of the candidate regulators
    if regulators == "all":
        input_idx = range(ngenes)
    else:
        input_idx = [i for i, gene in enumerate(gene_names) if gene in regulators]

    # Get the non-ranked list of regulatory links
    vInter = [
        (i, j, score) for (i, j), score in ndenumerate(VIM) if i in input_idx and i != j
    ]

    # Rank the list according to the weights of the edges
    vInter_sort = sorted(vInter, key=itemgetter(2), reverse=True)
    nInter = len(vInter_sort)

    # Random permutation of edges with score equal to 0
    flag = 1
    i = 0
    while flag and i < nInter:
        (TF_idx, target_idx, score) = vInter_sort[i]
        if score == 0:
            flag = 0
        else:
            i += 1

    if not flag:
        items_perm = vInter_sort[i:]
        items_perm = random.permutation(items_perm)
        vInter_sort[i:] = items_perm

    # Write the ranked list of edges
    nToWrite = nInter
    if isinstance(maxcount, int) and maxcount >= 0 and maxcount < nInter:
        nToWrite = maxcount

    if file_name:
        outfile = open(file_name, "w")

        if gene_names is not None:
            for i in range(nToWrite):
                (TF_idx, target_idx, score) = vInter_sort[i]
                TF_idx = int(TF_idx)
                target_idx = int(target_idx)
                outfile.write(
                    "%s\t%s\t%.6f\n"
                    % (gene_names[TF_idx], gene_names[target_idx], score)
                )
        else:
            for i in range(nToWrite):
                (TF_idx, target_idx, score) = vInter_sort[i]
                TF_idx = int(TF_idx)
                target_idx = int(target_idx)
                outfile.write("G%d\tG%d\t%.6f\n" % (TF_idx + 1, target_idx + 1, score))

        outfile.close()

    else:
        if gene_names is not None:
            for i in range(nToWrite):
                (TF_idx, target_idx, score) = vInter_sort[i]
                TF_idx = int(TF_idx)
                target_idx = int(target_idx)
                print(
                    "%s\t%s\t%.6f" % (gene_names[TF_idx], gene_names[target_idx], score)
                )
        else:
            for i in range(nToWrite):
                (TF_idx, target_idx, score) = vInter_sort[i]
                TF_idx = int(TF_idx)
                target_idx = int(target_idx)
                print("G%d\tG%d\t%.6f" % (TF_idx + 1, target_idx + 1, score))


def GENIE3(
    expr_data,
    gene_names=None,
    regulators="all",
    tree_method="RF",
    K="sqrt",
    ntrees=1000,
    nthreads=1,
):
    """Computation of tree-based scores for all putative regulatory links.

    Parameters
    ----------

    expr_data: numpy array
        Array containing gene expression values. Each row corresponds to a condition and each column corresponds to a gene.

    gene_names: list of strings, optional
        List of length p, where p is the number of columns in expr_data, containing the names of the genes. The i-th item of gene_names must correspond to the i-th column of expr_data.
        default: None

    regulators: list of strings, optional
        List containing the names of the candidate regulators. When a list of regulators is provided, the names of all the genes must be provided (in gene_names). When regulators is set to 'all', any gene can be a candidate regulator.
        default: 'all'

    tree-method: 'RF' or 'ET', optional
        Specifies which tree-based procedure is used: either Random Forest ('RF') or Extra-Trees ('ET')
        default: 'RF'

    K: 'sqrt', 'all' or a positive integer, optional
        Specifies the number of selected attributes at each node of one tree: either the square root of the number of candidate regulators ('sqrt'), the total number of candidate regulators ('all'), or any positive integer.
        default: 'sqrt'

    ntrees: positive integer, optional
        Specifies the number of trees grown in an ensemble.
        default: 1000

    nthreads: positive integer, optional
        Number of threads used for parallel computing
        default: 1


    Returns
    -------

    An array in which the element (i,j) is the score of the edge directed from the i-th gene to the j-th gene. All diagonal elements are set to zero (auto-regulations are not considered). When a list of candidate regulators is provided, the scores of all the edges directed from a gene that is not a candidate regulator are set to zero.

    """

    time_start = time.time()

    # Check input arguments
    if not isinstance(expr_data, ndarray):
        raise ValueError(
            "expr_data must be an array in which each row corresponds to a condition/sample and each column corresponds to a gene"
        )

    ngenes = expr_data.shape[1]

    if gene_names is not None:
        if not isinstance(gene_names, (list, tuple)):
            raise ValueError("input argument gene_names must be a list of gene names")
        elif len(gene_names) != ngenes:
            raise ValueError(
                "input argument gene_names must be a list of length p, where p is the number of columns/genes in the expr_data"
            )

    if regulators != "all":
        if not isinstance(regulators, (list, tuple)):
            raise ValueError("input argument regulators must be a list of gene names")

        if gene_names is None:
            raise ValueError(
                "the gene names must be specified (in input argument gene_names)"
            )
        else:
            sIntersection = set(gene_names).intersection(set(regulators))
            if not sIntersection:
                raise ValueError(
                    "the genes must contain at least one candidate regulator"
                )

    if tree_method != "RF" and tree_method != "ET":
        raise ValueError(
            'input argument tree_method must be "RF" (Random Forests) or "ET" (Extra-Trees)'
        )

    if K != "sqrt" and K != "all" and not isinstance(K, int):
        raise ValueError(
            'input argument K must be "sqrt", "all" or a stricly positive integer'
        )

    if isinstance(K, int) and K <= 0:
        raise ValueError(
            'input argument K must be "sqrt", "all" or a stricly positive integer'
        )

    if not isinstance(ntrees, int):
        raise ValueError("input argument ntrees must be a stricly positive integer")
    elif ntrees <= 0:
        raise ValueError("input argument ntrees must be a stricly positive integer")

    if not isinstance(nthreads, int):
        raise ValueError("input argument nthreads must be a stricly positive integer")
    elif nthreads <= 0:
        raise ValueError("input argument nthreads must be a stricly positive integer")

    print("Tree method: " + str(tree_method))
    print("K: " + str(K))
    print("Number of trees: " + str(ntrees))
    print("\n")

    # Get the indices of the candidate regulators
    if regulators == "all":
        input_idx = list(range(ngenes))
    else:
        input_idx = [i for i, gene in enumerate(gene_names) if gene in regulators]

    # Learn an ensemble of trees for each target gene, and compute scores for candidate regulators
    VIM = zeros((ngenes, ngenes))

    if nthreads > 1:
        print("running jobs on %d threads" % nthreads)

        input_data = list()
        for i in range(ngenes):
            input_data.append([expr_data, i, input_idx, tree_method, K, ntrees])

        pool = Pool(nthreads)
        alloutput = pool.map(wr_GENIE3_single, input_data)

        for i, vi in alloutput:
            VIM[i, :] = vi

    else:
        print("running single threaded jobs")
        for i in range(ngenes):
            print("Gene %d/%d..." % (i + 1, ngenes))

            vi = GENIE3_single(expr_data, i, input_idx, tree_method, K, ntrees)
            VIM[i, :] = vi

    VIM = transpose(VIM)

    time_end = time.time()
    print("Elapsed time: %.2f seconds" % (time_end - time_start))

    return VIM


def wr_GENIE3_single(args):
    return [
        args[1],
        GENIE3_single(args[0], args[1], args[2], args[3], args[4], args[5]),
    ]


def GENIE3_single(expr_data, output_idx, input_idx, tree_method, K, ntrees):
    ngenes = expr_data.shape[1]

    # Expression of target gene
    output = expr_data[:, output_idx]

    # Normalize output data
    output = output / std(output)

    # Remove target gene from candidate regulators
    input_idx = input_idx[:]
    if output_idx in input_idx:
        input_idx.remove(output_idx)

    expr_data_input = expr_data[:, input_idx]

    # Parameter K of the tree-based method
    if (K == "all") or (isinstance(K, int) and K >= len(input_idx)):
        max_features = "auto"
    else:
        max_features = K

    if tree_method == "RF":
        treeEstimator = RandomForestRegressor(
            n_estimators=ntrees, max_features=max_features
        )
    elif tree_method == "ET":
        treeEstimator = ExtraTreesRegressor(
            n_estimators=ntrees, max_features=max_features
        )

    # Learn ensemble of trees
    treeEstimator.fit(expr_data_input, output)

    # Compute importance scores
    feature_importances = compute_feature_importances(treeEstimator)
    vi = zeros(ngenes)
    vi[input_idx] = feature_importances

    return vi



import os
from pathlib import Path
from typing import Dict

# Local imports
from . import run_experiment as runexp
from . import post_processing as po


class BoolODE(object):
    """
    The BoolODE object is created by parsing a user-provided configuration
    file. It then calls the correct functions for executing the user defined
    experiment.
    """

    def __init__(
        self,
        jobs,  # joblist {"name": "dyn-LL-1", "simulation_time": 9, "num_cells": 300, "do_parallel": True, "sample_cells": False, 'nClusters': 1}
        dropout_jobs: Dict[
            str, str
        ] = None,  # Dropouts {"drop_cutoff": 0.5, "drop_prob": 0.4, "sample_size": 100}
        perplexities=None,  #
        slingshot_jobs=None,  # Slingshot {"perplexity": 200}
        gensample_jobs=None,  # GenSamples {"sample_size":100, "nDatasets":1}
        model_dir="data",
        output_dir="output",
        do_simulations=True,
        do_post_processing=True,
        modeltype="hill",
    ) -> None:
        """
            __init__ _summary_

            Args:
                jobs (_type_): _description_
                output_dir (_type_): _description_
                do_simulations (_type_): _description_
                do_post_processing (_type_): _description_
                modeltype (_type_): _description_
                nCells
        drop_cutoff
        drop_prob
        """
        self.dropout_jobs = dropout_jobs
        self.perplexities = perplexities
        self.slingshot_jobs = slingshot_jobs
        self.gensample_jobs = gensample_jobs
        self.geneexpression_jobs = None
        self.model_dir = model_dir
        self.output_dir = output_dir
        self.do_simulations = do_simulations
        self.do_post_processing = do_post_processing
        self.modeltype = modeltype
        self.jobs = self._process_jobs([jobs])

    def _process_jobs(self, jobs) -> Dict[int, Dict]:
        """
        Creates a list of jobs, where each job is specified by
        the values in a dictionary called data.
        Default parameter values are specified here.
        """
        n_jobs = {}
        for jobid, job in enumerate(jobs):
            data = {}
            # Create output folder if it doesnt exist
            data["name"] = job.get("name")
            data["outprefix"] = Path(self.output_dir, job.get("name"))
            data["modelpath"] = Path(self.model_dir, job.get("model_definition"))
            data["simulation_time"] = job.get("simulation_time", 20)
            data["icsPath"] = Path(
                self.model_dir, job.get("model_initial_conditions", "")
            )
            data["num_cells"] = job.get("num_cells", 100)
            data["sample_cells"] = job.get("sample_cells", False)
            data["nClusters"] = job.get("nClusters", 1)
            data["doParallel"] = job.get("do_parallel", False)
            data["identical_pars"] = job.get("identical_pars", False)
            data["sample_pars"] = job.get("sample_pars", False)
            data["sample_std"] = job.get("sample_std", 0.1)
            data["integration_step_size"] = job.get("integration_step_size", 0.01)
            # Optional Settings
            data["parameter_inputs_path"] = Path(
                self.model_dir, job.get("parameter_inputs_path", "")
            )
            data["parameter_set"] = Path(self.model_dir, job.get("parameter_set", ""))
            data["interaction_strengths"] = Path(
                self.model_dir, job.get("interaction_strengths", "")
            )
            data["species_type"] = Path(self.model_dir, job.get("species_type", ""))
            # Simulator settings
            data["burnin"] = job.get("burn_in", False)
            data["writeProtein"] = job.get("write_protein", False)
            data["normalizeTrajectory"] = job.get("normalize_trajectory", False)
            data["add_dummy"] = job.get("add_dummy", False)
            data["max_parents"] = job.get("max_parents", 1)
            data["modeltype"] = self.modeltype

            n_jobs[jobid] = data
        return n_jobs

    def execute_jobs(self, parallel=False, num_threads=1):
        """
        Run each user specified job.
        BoolODE runs two types of functions
        1. If `do_simulation == TRUE`, perform SDE simulations of model specified as Boolean rules.
        2. If `do_post_processing == TRUE` perform the list of post processing operations specified.

        .. warning::
            This function automatically creates folders for each job name
            as specified in the config file, if the folder doesn't already exist.
            Contents of existing folders will be rewritten!
        """
        base_output_dir = self.output_dir

        alljobs = self.jobs.keys()
        print("Creating output folders")
        for jobid in alljobs:
            outdir = self.jobs[jobid]["outprefix"]
            if not os.path.exists(outdir):
                print(outdir, "does not exist, creating it...")
                os.makedirs(outdir)
        if self.do_simulations:
            print("Starting simulations")
            for jobid in alljobs:
                runexp.startRun(self.jobs[jobid])
        if self.do_post_processing:
            print("Starting post processing")
            self.do_post_processing()

    def do_post_processing(self):
        """
        Call genSamples() first. Then run DimRed runSlingShot,
        generateDropouts, if specified by the user.
        """
        alljobs = list(self.jobs.keys())
        filetypedict = {
            "expr": "ExpressionData.csv",
            "pseudo": "PseudoTime.csv",
            "refNet": "refNetwork.csv",
        }

        ## Always do genSamples() once if even a single other analysis is requested!
        doOtherAnalysis = False
        if (
            self.dropout_jobs
            or self.perplexities
            or self.geneexpression_jobs
            or self.slingshot_jobs
        ):
            doOtherAnalysis = True
        generatedPaths = {}

        if self.gensample_jobs is not None or doOtherAnalysis:
            print("Generating Samples...")
            if self.gensample_jobs is None:
                gsamp = {}
                gsamp["sample_size"] = self.jobs[alljobs[0]]["num_cells"]
                gsamp["nDatasets"] = 1
                gensample_jobs = [gsamp]
            else:
                gensample_jobs = self.gensample_jobs
            for gsamp in gensample_jobs:
                for jobid in alljobs:
                    settings = {}
                    settings["num_cells"] = self.jobs[jobid]["num_cells"]
                    settings["sample_size"] = gsamp.get("sample_size", 100)
                    settings["outPrefix"] = str(self.jobs[jobid]["outprefix"])
                    settings["nDatasets"] = gsamp.get("nDatasets", 1)
                    settings["name"] = self.jobs[jobid]["name"]
                    settings["nClusters"] = self.jobs[jobid]["nClusters"]
                    generatedPaths[jobid] = po.genSamples(settings)

        if self.dropout_jobs is not None:
            print("Starting genDropouts...")
            for drop in self.dropout_jobs:
                num_invalid = 0
                for jobid in alljobs:
                    for gsampPath in generatedPaths[jobid]:
                        settings = {}
                        invalid = False
                        settings["outPrefix"] = gsampPath
                        settings["expr"] = Path(gsampPath, "ExpressionData.csv")
                        settings["pseudo"] = Path(gsampPath, "PseudoTime.csv")
                        settings["refNet"] = Path(gsampPath, "refNetwork.csv")
                        settings["dropout"] = drop.get("dropout", True)
                        settings["sample_size"] = drop.get("sample_size", 100)
                        settings["num_cells"] = self.jobs[jobid]["num_cells"]
                        settings["drop_cutoff"] = drop.get("drop_cutoff", 0.0)
                        settings["drop_prob"] = drop.get("drop_prob", 0.0)

                        for filetype in ["expr", "pseudo", "refNet"]:
                            if not settings[filetype].is_file():
                                print(
                                    self.jobs[jobid]["name"],
                                    ": ",
                                    filetypedict[filetype],
                                    "not found. Retry with `do_simulations: True` in global_settings.",
                                )
                                invalid = True
                                num_invalid += 1
                                break
                        if not invalid:
                            po.genDropouts(settings)
                    if num_invalid == len(alljobs):
                        break

        if self.perplexities is not None:
            print("Starting dimesionality reduction using tSNE")
            for perplexity in self.perplexities:
                num_invalid = 0
                for jobid in alljobs:
                    for gsampPath in generatedPaths[jobid]:
                        print(f"perplexity=", perplexity)
                        settings = {}
                        invalid = False
                        settings["expr"] = Path(gsampPath, "ExpressionData.csv")
                        settings["pseudo"] = Path(gsampPath, "PseudoTime.csv")
                        settings["perplexity"] = perplexity
                        settings["default"] = False
                        for filetype in ["expr", "pseudo"]:
                            if not settings[filetype].is_file():
                                print(
                                    self.jobs[jobid]["name"],
                                    ":",
                                    filetypedict[filetype],
                                    "not found. Retry with `do_simulations: True` in global_settings.",
                                )
                                invalid = True
                                num_invalid += 1
                                break
                        if not invalid:
                            po.doDimRed(settings)
                            all_invalid = False
                    if num_invalid == len(alljobs):
                        break

        if self.geneexpression_jobs is not None:
            if self.perplexities is None:
                print("Using default perplexity=50 (Specify `perplexity` under DimRed)")
                perplexity = 50
            else:
                if len(self.perplexities) > 1:
                    perplexity = min([j["perplexity"] for j in self.perplexities])
                else:
                    perplexity = self.perplexities[0]["perplexity"]

            print("Plotting gene expression levels in tSNE projection")
            for jobid in alljobs:
                for gsampPath in generatedPaths[jobid]:
                    print(gsampPath)
                    settings = {}
                    invalid = False
                    settings["expr"] = Path(gsampPath, "ExpressionData.csv")
                    settings["pseudo"] = Path(gsampPath, "PseudoTime.csv")
                    settings["perplexity"] = perplexity
                    settings["default"] = False
                    po.plotGeneExpression(settings)

        if self.slingshot_jobs is not None:
            if self.perplexities is None:
                print(
                    "Using default perplexity=300. (Specify `perplexity` under DimRed.)"
                )
            print("Starting SlingShot...")
            for sshot in self.slingshot_jobs:
                for jobid in alljobs:
                    for gsampPath in generatedPaths[jobid]:
                        settings = {}
                        invalid = False
                        settings["outPrefix"] = (
                            gsampPath + "/" + gsampPath.split("/")[-1] + "-ss"
                        )
                        settings["expr"] = Path(gsampPath, "ExpressionData.csv")
                        settings["pseudo"] = Path(gsampPath, "PseudoTime.csv")
                        settings["refNet"] = Path(gsampPath, "refNetwork.csv")
                        if self.jobs[jobid]["nClusters"] == 1:
                            settings["nClusters"] = 1
                        else:
                            settings["nClusters"] = self.jobs[jobid]["nClusters"] + 1
                        settings["noEnd"] = sshot.get("noEnd", False)
                        settings["perplexity"] = sshot.get("perplexity", 300)

                        for filetype in ["expr", "pseudo", "refNet"]:
                            if not settings[filetype].is_file():
                                print(
                                    self.jobs[jobid]["name"],
                                    ": ",
                                    filetypedict[filetype],
                                    "not found. Retry with `do_simulations: True` in global_settings.",
                                )
                                invalid = True
                                break
                        if not invalid:
                            po.computeSSPT(settings)



import os
from tqdm import tqdm
import numpy as np
import pandas as pd
from pathlib import Path
from sklearn.cluster import KMeans
from sklearn.manifold import TSNE
import matplotlib.pyplot as plt


def genSamples(opts):
    """
    Generate samples of cells from a given set of simulations. This sample
    will then be used for other post processing steps.
    """
    numclusters = opts["nClusters"]
    num_simulations = opts["num_cells"]

    if numclusters > 1:
        clusterdf = pd.read_csv(opts["outPrefix"] + "/ClusterIds.csv", index_col=0)

    sample_size = opts["sample_size"]
    if opts["sample_size"] > num_simulations:
        print("sample_size should be less than num of experiments")
        sample_size = num_simulations

    df = pd.read_csv(opts["outPrefix"] + "/simulations/E0.csv", index_col=0)
    maxtime = len(df.columns)

    generatedPaths = []
    for did in range(1, opts["nDatasets"] + 1):
        # example:
        # Beeline/inputs/DYN-LI-500-1/...
        outfpath = (
            opts["outPrefix"]
            + "/"
            + opts["name"]
            + "-"
            + str(sample_size)
            + "-"
            + str(did)
        )
        generatedPaths.append(outfpath)

        if not os.path.exists(outfpath):
            print(outfpath, "does not exist, creating it...")
            os.makedirs(outfpath)
        # Create cell ids
        simids = np.random.choice(
            range(num_simulations), size=sample_size, replace=False
        )
        fids = ["E" + str(sid) + ".csv" for sid in simids]
        timepoints = np.random.choice(range(1, maxtime), size=sample_size)
        min_t = min(timepoints)
        max_t = max(timepoints)
        pts = [(t - min_t) / (max_t - min_t) for t in timepoints]
        cellids = ["E" + str(sid) + "_" + str(t) for sid, t in zip(simids, timepoints)]
        # Read simulations from input dataset #psetid
        # to build a sample
        sample = []
        for fid, cid in tqdm(zip(fids, cellids)):
            df = pd.read_csv(opts["outPrefix"] + "/simulations/" + fid, index_col=0)
            df.sort_index(inplace=True)
            sample.append(df[cid].to_frame())
        sampledf = pd.concat(sample, axis=1)
        sampledf.to_csv(outfpath + "/ExpressionData.csv")
        ## Read refNetwork.csv
        refdf = pd.read_csv(opts["outPrefix"] + "/refNetwork.csv")
        refdf.to_csv(outfpath + "/refNetwork.csv", index=False)
        if numclusters == 1:
            ptdf = pd.DataFrame(
                np.array(pts), index=pd.Index(cellids), columns=["PseudoTime"]
            )
        else:
            subsetcluster = clusterdf.loc[[cid.split("_")[0] for cid in cellids]]
            ptdf = pd.DataFrame(
                index=pd.Index(cellids),
                columns=["PseudoTime" + str(1 + i) for i in range(numclusters)],
            )
            for (cid, row), pt in zip(ptdf.iterrows(), pts):
                ptdf.loc[cid][
                    "PseudoTime"
                    + str(int(subsetcluster.loc[cid.split("_")[0]]["cl"]) + 1)
                ] = round(pt, 5)
        ptdf.to_csv(outfpath + "/PseudoTime.csv", na_rep="NA")
    return generatedPaths


def genDropouts(opts):
    """Induce drop out in the simulated scRNAseq datasets."""
    if opts["dropout"]:
        dropoutCutoffs = opts["drop_cutoff"]
    else:
        dropoutCutoffs = 0

    ## Read the ExpressionData.csv file
    expDF = pd.read_csv(opts["expr"], index_col=0)
    ## Read the PseudoTime.csv file
    PTDF = pd.read_csv(opts["pseudo"], index_col=0)
    ## Read the refNetwork.csv file
    refDF = pd.read_csv(opts["refNet"], index_col=0)

    ## Generate output path for the dropout datasets
    path = opts["outPrefix"]  # +str(ncells)
    path += "-" + str(int(100 * dropoutCutoffs)) + "-" + str(opts["drop_prob"])
    if not os.path.exists(path):
        os.makedirs(path)

    # Sample cells
    DropOutDF = expDF.copy()
    # DropOutDF = DropOutDF.sample(n=ncells, axis='columns')
    # cellIDs = DropOutDF.columns

    # copy over PT and refNetwork files
    refDF.to_csv(path + "/refNetwork.csv")
    PTDF.to_csv(path + "/PseudoTime.csv")

    # Drop-out genes if they are less than the
    # percentile value @ "dc" with 50% chance
    if dropoutCutoffs != 0:
        quantileExp = expDF.quantile(q=dropoutCutoffs, axis="columns")
        for idx, row in tqdm(expDF.iterrows()):
            for col in expDF.columns:
                if row[col] < quantileExp.loc[idx]:
                    cointoss = np.random.random()
                    if cointoss < opts["drop_prob"]:
                        DropOutDF.loc[idx, col] = 0.0

    DropOutDF.to_csv(path + "/ExpressionData.csv")


def doDimRed(opts):
    """
    Carry out dimensionality reduction
    """
    ExpDF = pd.read_csv(opts["expr"], index_col=0, header=0)
    ptDF = pd.read_csv(opts["pseudo"], index_col=0, header=0)
    perplexity = opts["perplexity"]
    print(perplexity)
    print("Computing TSNE...")
    DimRedRes = TSNE(n_components=2, perplexity=perplexity).fit_transform(ExpDF.T)

    DimRedDF = pd.DataFrame(
        DimRedRes, columns=["dim1", "dim2"], index=pd.Index(list(ExpDF.columns))
    )
    DimRedDF.loc[:, "pt"] = ptDF.min(axis="columns")
    DimRedDF.to_csv(
        str(opts["expr"].parent) + "/tsne" + str(perplexity) + ".tsv", sep="\t"
    )
    plt.figure()
    plt.scatter(DimRedDF.dim1, DimRedDF.dim2, c=DimRedDF.pt)
    plt.savefig(str(opts["expr"].parent) + "/tsne" + str(perplexity) + ".png")
    plt.close()


def computeSSPT(opts):
    """
    *Author: Aditya Pratapa*

    Compute PseudoTime using 'slingshot'.
    Needs the input Gene x Cells expression data frame.
    Needs number of clusters to be expected in the
    dataset.
    E.g., Linear: k=1 (no PseudoTime inference is done)
    E.g., Bifurcating: k=3 (1 initial and 2 terminal)
    E.g., Trifurcating: k=4 (1 initial and 3 terminal)
    """
    ExpDF = pd.read_csv(opts["expr"], index_col=0, header=0)
    ptDF = pd.read_csv(opts["pseudo"], index_col=0, header=0)
    nClust = opts["nClusters"]
    outPath = opts["outPrefix"]
    perplexity = opts["perplexity"]
    noEnd = opts["noEnd"]

    if nClust == 1:
        # Return simulation time as PseduoTime
        ptDF.loc[ExpDF.columns].to_csv(outPath + "/PseudoTime.csv", columns=["Time"])

    else:
        ### Compute PseudoTime ordering using slingshot

        # Step-1: Compute dimensionality reduction. This is done in doDimRed().
        # Currently only does TSNE
        # TODO: Add PCA

        # Step-2: Read TSNE results to a dataframe
        if not Path(settings["expr"].parent / "tsne.tsv").is_file():
            # First compute tSNE if this hasn't been done
            doDimRed(opts)
        DimRedDF = pd.read_csv(settings["expr"].parent / "tsne.tsv", sep="\t")

        # Step-3: Compute kMeans clustering
        DimRedDF.loc[:, "cl"] = KMeans(n_clusters=nClust).fit(ExpDF.T).labels_
        # Identify starting cluster

        # Step-4: Identify cells corresponding to initial and final states
        # Cells in initial states are identified from cluster with smallest
        # mean experimental time
        # Cells in final states are rest of the clusters
        startClust = DimRedDF.groupby("cl").mean()["pt"].idxmin()
        endClust = ",".join(
            [
                str(ix)
                for ix in DimRedDF.groupby("cl").mean()["pt"].index
                if ix != startClust
            ]
        )
        startClust = str(startClust)

        # Step-5: Create a temporary directory and write files necessary to run slingshot
        os.makedirs(outPath, exist_ok=True)

        DimRedDF.to_csv(outPath + "/rd.tsv", columns=["dim1", "dim2"], sep="\t")
        DimRedDF.to_csv(outPath + "/cl.tsv", columns=["cl"], sep="\t")
        if noEnd:
            cmdToRun = " ".join(
                [
                    "docker run --rm -v",
                    str(Path.cwd()) + "/" + outPath + "/:/data/temp",
                    'slingshot:base /bin/sh -c "Rscript data/run_slingshot.R',
                    "--input=/data/temp/rd.tsv --input-type=matrix",
                    "--cluster-labels=/data/temp/cl.tsv",
                    "--start-clus=" + startClust + '"',
                ]
            )

        else:
            cmdToRun = " ".join(
                [
                    "docker run --rm -v",
                    str(Path.cwd()) + "/" + outPath + "/:/data/temp",
                    'slingshot:base /bin/sh -c "Rscript data/run_slingshot.R',
                    "--input=/data/temp/rd.tsv --input-type=matrix",
                    "--cluster-labels=/data/temp/cl.tsv",
                    "--start-clus=" + startClust,
                    "--end-clus=" + endClust + '"',
                ]
            )
        print(cmdToRun)
        os.system(cmdToRun)
        # Do this only for the first file
        tn = pd.read_csv(outPath + "/rd.tsv", header=0, index_col=None, sep="\t")

        cl = pd.read_csv(outPath + "/cl.tsv", header=0, index_col=0, sep="\t")

        curveFile = open(outPath + "/curves.csv", "r")
        curveLst = []
        lneCnt = 0
        for line in curveFile:
            curveLst.append([float(p) for p in line.strip().split(",")])
            lneCnt += 1

        tn["cl"] = cl.values
        tn.columns = ["CellID", "dim1", "dim2", "kMeans"]
        tn.index = tn.CellID

        f, axes = plt.subplots(2, 2, figsize=(7.5, 7.5))

        # Plot slingshot pseudotime
        # and original clusters
        detPT = pd.read_csv(outPath + "/SlingshotPT.csv", header=0, index_col=0)
        print()
        colNames = detPT.columns
        for colName in colNames:
            # Select cells belonging to each pseudotime trajectory
            index = detPT[colName].index[detPT[colName].notnull()]
            tn.loc[index, colName] = detPT.loc[index, colName]

            sns.scatterplot(
                x="dim1",
                y="dim2",
                data=tn.loc[index],
                hue=colName,
                palette="viridis",
                ax=axes[1][0],
            )
            plt.legend([])

        for line in range(0, lneCnt, 2):
            sns.lineplot(
                x=curveLst[line + 1], y=curveLst[line], color="k", ax=axes[1][0]
            )

        sns.scatterplot(
            x="dim1", y="dim2", data=tn, hue="kMeans", palette="Set1", ax=axes[1][1]
        )

        # Plot deterministic pseduotime
        # and original clusters
        detPT = pd.read_csv(outPath + "/PseudoTime.csv", header=0, index_col=0)
        colNames = detPT.columns
        for idx in range(len(colNames)):
            # Select cells belonging to each pseudotime trajectory
            colName = colNames[idx]
            index = detPT[colName].index[detPT[colName].notnull()]
            tn.loc[index, "Original"] = int(idx)

        tn["ExpTime"] = detPT.min(axis="columns")

        sns.scatterplot(
            x="dim1", y="dim2", data=tn, hue="ExpTime", palette="viridis", ax=axes[0][0]
        )

        sns.scatterplot(
            x="dim1", y="dim2", data=tn, hue="Original", palette="Set1", ax=axes[0][1]
        )

        axes[0][0].get_legend().remove()
        axes[0][0].title.set_text("Experiment Time")
        axes[0][1].get_legend().remove()
        axes[0][1].title.set_text("Original Trajectories")

        axes[1][0].get_legend().remove()
        axes[1][0].title.set_text("Slingshot Pseudotime")

        axes[1][1].get_legend().remove()
        axes[1][1].title.set_text("kMeans Clustering")

        f.tight_layout()

        tn.to_csv(outPath + "/Updated_rd.tsv", sep="\t")
        plt.savefig(outPath + "/SlingshotOutput.png")
    os.system("rm -rf temp/")



#!/usr/bin/env python
# coding: utf-8
import numpy as np
import pandas as pd

from itertools import combinations

# local imports
from . import utils


class GenerateModel:
    """Class that holds model attributes. Provides helper functions to convert a Boolean model
    to an ODE model

    :param settings: User defined job settings. This is created in boolode.do_post_processing()
    :type settings: dict
    :param parameterInputsDF: User specified parmeter inputs. Specified in config YAML as a path.
    :type parameterInputsDF: pandas DataFrame
    :param parameterSetDF: User specified kinetic parameters. Specified in config YAML as a path.
    :type parameterSetDF: pandas DataFrame
    :param interactionStrengthDF: User specified interaction strengths. Specified in config YAML as a path
    :type interactionStrengthDF: pandas DataFrame
    """

    def __init__(
        self, settings, parameterInputsDF, parameterSetDF, interactionStrengthDF
    ) -> None:
        """
        This class is initialized with the job_settings dictionary created
        by the BoolODE class. Additionally, three dataframe objects are required:

        1. parameterInputsDF - specifies parameter inputs
        2. parameterSetDF - specifies pregenerated python file containing kinetic parameter values
        3. interactionStrengthDF - specifies interactions strengths

        By default, if these aren't specified by the user, BoolODE stores empty pandas DataFrame objects.
        """
        self.settings = settings
        self.parameterInputsDF = parameterInputsDF
        self.parameterSetDF = parameterSetDF
        self.interactionStrengthDF = interactionStrengthDF
        # Class variables
        self.withRules = list()
        self.withoutRules = list()
        self.allnodes = set()
        self.varspecs = dict()
        self.varmapper = dict()
        self.par = dict()
        self.parmapper = dict()
        self.kineticParameterDefaults = dict()
        self.inputs = list()
        self.genelist = list()
        self.proteinlist = list()
        self.nodeTypeDF = pd.DataFrame()
        self.df = pd.DataFrame()
        self.ModelSpec = dict()
        self.path_to_ode_model = str()
        # Read the model definition
        # 1. populate self.df
        # 2. store genelist, withRules, withoutRules, allnodes
        # 3. initialize varspecs
        self.readBooleanRules()
        # Create model parameters and assign values
        self.getParameters()
        # Create the model dictionary
        self.generateModelDict()
        # Write ODE model to file
        self.writeModelToFile()
        # Write parameters to file
        self.writeParametersToFile()

    def readBooleanRules(self):
        """
        Reads a rule file from path and stores the rules in a dictionary.
        Performs the following transformations:

        1. Parses each Boolean rule to extract nodes in the model
        2. Identifies nodes without corresponding rules: (a) If a parameterInput file is specified and the node under consideration
           is present in the file, the node is treated as an input parameter
           (b) Else, a self-edge is added to the node
        3. Adds dummy genes - *This is experimental*
        4. Finally, create a gene (x\_) and protein (p\_)  variable corresponding to
           each node in the network, taking into consideration user defined types.
        """
        self.df = pd.read_csv(self.settings["modelpath"], sep="\t", engine="python")

        self.withRules = list(self.df["Gene"].values)

        ## Extract all nodes from the boolean rules
        ## Note that not all nodes might have a rule attached to them
        for ind, row in self.df.iterrows():
            rhs = row["Rule"]
            rhs = rhs.replace("(", " ")
            rhs = rhs.replace(")", " ")
            tokens = rhs.split(" ")
            reg = [t for t in tokens if t not in ["not", "and", "or", ""]]
            self.allnodes.update(set(reg))

        self.withoutRules = list(self.allnodes.difference(set(self.withRules)))

        ## Every node without a rule is treated as follows:
        ## If the user has specified a Parameter Input file treat as parameter, else
        for n in self.withoutRules:
            if (
                not self.parameterInputsDF.empty
                and n in self.parameterInputsDF["Input"]
            ):
                print("Treating %s as parameter" % {n})
            else:
                print(n, "has no rule, adding self-activation.")
                self.df = self.df.append({"Gene": n, "Rule": n}, ignore_index=True)
                self.withRules.append(n)
                self.withoutRules.remove(n)

        if self.settings["add_dummy"]:
            self.addDummyGenes()

        # Variables:
        ## Check:
        ## If nodeType is provided, initialize the
        ## correct node names.
        ## Rule:
        ## Every gene (x_node) gets a corresponding
        ## protein (p_node). The gene equation contains
        ## regulatory logic, protein equation contains production
        ## degradation terms.
        ## Every 'protein' (p_node) only gets one variable,
        ## and the corresponding equation contains the regulatory
        ## logic term. There is no production/degradation associated
        ## with this node.

        # Create empty equations
        if self.nodeTypeDF.empty:
            # Assume everything is a gene, so make the corresponding protein
            for node in self.withRules:
                self.varspecs["x_" + node] = ""
                self.varspecs["p_" + node] = ""
                self.genelist.append(node)
        else:
            # TODO Consider removing support for node type.
            for i, row in self.nodeTypeDF.iterrows():
                node = row["Node"]
                if row["Type"] == "protein":
                    self.varspecs.update({"p_" + node: ""})
                    self.proteinlist.append(node)
                elif row["Type"] == "gene":
                    self.varspecs.update({"x_" + node: "", "p_" + node: ""})
                    self.genelist.append(node)
                else:
                    print(row["Type"], "not recognized. Treating as gene.")
                    self.varspecs.update({"x_" + node: "", "p_" + node: ""})
                    self.genelist.append(node)

    def addDummyGenes(self):
        """
        Add dummy genes.
        Starts with the user specified Boolean model, and grows the
        network by adding new 'dummy' nodes to each existing node.
        In the current configuration, the user can specify `max_parents`
        which specifies the number of parent nodes of every dummy node.
        Currently, twice as many dummy nodes as nodes in the original graph
        are added.

        .. todo::
            Expand this function by adding user defined number of nodes.

        .. warning::
            This feature is experimental.
        """
        print("using max_parents=" + str(self.settings["max_parents"]))
        num_dummy = 2 * len(allnodes)
        for dg in range(num_dummy):
            ## Variable number of parents
            # DF = DF.append({'Gene':'dummy' + str(dg),
            #                 'Rule':' or '.join([s for s in \
            #                                     np.random.choice(list(allnodes),
            #                                                      size=np.random.choice([i+1 for i in range(len(allnodes)-1)]))])},
            #                ignore_index = True)
            self.df = self.df.append(
                {
                    "Gene": "dummy" + str(dg),
                    "Rule": " or ".join(
                        [s for s in np.random.choice(list(allnodes), size=max_parents)]
                    ),
                },
                ignore_index=True,
            )
        self.df.to_csv(self.setting["outPrefix"] + "rules-with-added-genes.csv")

    def getParameters(self):
        """Create and assigns the kinetic parameters for each equations the
        corresponding parameter value.

        .. note::
            Default kinetic parameter values are stored in ./parameters.yaml. This
            is provided to make it easy for users to define different defaults, but it
            other parameter values have not been tested extensively, and might not
            yield the expected results.
        """

        self.kineticParameterDefaults = utils.loadParameterValues()

        ## Derived parameters:
        ## The threshold is calculated as the half max value of the species
        x_max = (
            self.kineticParameterDefaults["mRNATranscription"]
            / self.kineticParameterDefaults["mRNADegradation"]
        )
        y_max = x_max * (
            self.kineticParameterDefaults["proteinTranslation"]
            / self.kineticParameterDefaults["proteinDegradation"]
        )

        hillThreshold = y_max / 2
        heavisideOmega = 2.0 / y_max

        self.kineticParameterDefaults["x_max"] = x_max
        self.kineticParameterDefaults["y_max"] = y_max
        self.kineticParameterDefaults["hillThreshold"] = hillThreshold
        self.kineticParameterDefaults["heavisideOmega"] = heavisideOmega

        parameterNamePrefixAndDefaultsAll = {
            # Hill coefficients
            "n_": self.kineticParameterDefaults["hillCoefficient"],
            # Thresholds
            "k_": hillThreshold,
            # Soft-Heaviside function steepness
            "sigmaH_": self.kineticParameterDefaults["heavisideSigma"],
        }
        parameterNamePrefixAndDefaultsGenes = {
            # mRNA transcription rates
            "m_": self.kineticParameterDefaults["mRNATranscription"],
            # mRNA degradation rates
            "l_x_": self.kineticParameterDefaults["mRNADegradation"],
            # Protein translation rate
            "r_": self.kineticParameterDefaults["proteinTranslation"],
            # protein degradation rates
            "l_p_": self.kineticParameterDefaults["proteinDegradation"],
        }

        par = dict()
        ## If there is even one signaling protein,
        ## create the y_max parameter
        if len(self.proteinlist) > 0:
            self.par["y_max"] = y_max
            self.par["signalingtimescale"] = signalingtimescale

        interactionStrengths = {}

        ## Carry out a series of checks to handle user defined input to model

        ## Check 1: handle "input parameters"
        if not self.parameterInputsDF.empty:
            # TODO The name 'parameter inputs' is confusing/misleading, considering renaming?
            self.addParameterInputs()

        ## Check 2: include user-specified interaction strengths
        if not self.interactionStrengthDF.empty:
            self.addInteractionStrengths(hillThreshold)

        ## Assigining values to each kinetic parameter
        ## Create a parameter of each parameter type for each gene.
        ## If samplePars=True, sample values from normal distributions,
        ## else, set all parameters to defaults
        if self.settings["sample_pars"]:
            self.assignSampledParameterValues(
                parameterNamePrefixAndDefaultsAll, parameterNamePrefixAndDefaultsGenes
            )
        else:
            self.assignDefaultParameterValues(
                parameterNamePrefixAndDefaultsAll, parameterNamePrefixAndDefaultsGenes
            )

        # Final check:
        # Update user defined parameter values
        if not self.parameterSetDF.empty:
            for pname, pval in self.parameterSetDF.iterrows():
                self.par[pname] = pvalue[1]

    def addParameterInputs(self):
        """
        Check for user specified input parameters.
        If a parameter input file is specified,
        Every row in parameterInputsDF contains two
        columns:

        1. 'Input' specifies the name of input parameter
        2. 'Value' specified the input state, typically 1 or 0, but for HIGH and LOW inputs respectively, but any float between 0 and 1 is valid input.

        Now, we create a fake hill term for this input, even
        though there is no equation corresponding to it. Thus,
        we create a hillThreshold and hillCoefficient
        term for it. This is useful so we can treat this input
        parameter as a "regulator" and leave the structure of
        the logic term unchanged.

        .. todo::
            Add appropriate parameters if heaviside
        """
        parameterInputsDict = {}
        hillThreshold = self.kineticParameterDefaults["hillThreshold"]
        hillCoefficient = self.kineticParameterDefaults["hillCoefficient"]
        for i, row in self.parameterInputsDF.iterrows():
            # initialize
            # parameterInputsDict[i] = {p:0 for p in self.withoutRules} ## Suspicious
            inputParam = row["Input"]
            if inputParam not in self.withoutRules:
                print(inputParam, "not found nodes without rules. Skipping")
            else:
                ## This can take values between 0 and 1
                if row["Value"] > 1.0:
                    inputValue = 1.0
                elif row["Value"] < 0.0:
                    inputValue = 0.0
                else:
                    inputValue = row["Value"]
                # The following line sets the appropriate constant input
                # Note: 2*hillThreshold is the max value that a protein can take
                parameterInputsDict[inputParam] = inputValue * 2 * hillThreshold
        # Add input parameters, set to the value calculated above
        self.par.update({p: v for p, v in parameterInputsDict.items()})
        if self.settings["modeltype"] == "hill":
            self.par.update(
                {"k_" + p: hillThreshold for p in parameterInputsDict.keys()}
            )
            self.par.update(
                {"n_" + p: hillCoefficient for p in parameterInputsDict.keys()}
            )

    def addInteractionStrengths(self, hillThreshold):
        """Check if interaction strengths are specified.
        Create a new parameter specifying the strength
        of a given interaction. For all other equations in
        which "regulator" appears in, its hillThreshold value
        will remain the default value.

        .. todo::
            Add appropriate parameters if heaviside
        """
        for i, row in self.interactionStrengthDF.iterrows():
            regulator = row["Gene2"]
            target = row["Gene1"]
            interactionStrength = row["Strength"]
            if self.settings["modeltype"] == "hill":
                self.par.update(
                    {
                        "k_"
                        + regulator
                        + "_"
                        + target: hillThreshold / interactionStrength
                    }
                )
            elif self.settings["modeltype"] == "heaviside":
                self.par.update(
                    {
                        "w_"
                        + regulator
                        + "_"
                        + target: hillThreshold / interactionStrength
                    }
                )

    def assignDefaultParameterValues(
        self, parameterNamePrefixAndDefaultsAll, parameterNamePrefixAndDefaultsGenes
    ):
        """
        Set each kinetic parameter to its default value.
        """
        print("Fixing rate parameters to defaults")
        for parPrefix, parDefault in parameterNamePrefixAndDefaultsAll.items():
            for node in self.withRules:
                self.par[parPrefix + node] = parDefault

        for parPrefix, parDefault in parameterNamePrefixAndDefaultsGenes.items():
            for node in self.withRules:
                if node in self.genelist:
                    self.par[parPrefix + node] = parDefault

    def assignSampledParameterValues(
        self, parameterNamePrefixAndDefaultsAll, parameterNamePrefixAndDefaultsGenes
    ):
        """
        Sample kinetic parameters from a truncated normal distribution with mean=default
        parameter value, and standard deviation = sample_std, in a range +- 10% of
        the default value.
        """
        print("Sampling parameter values")
        print("Using std=" + str(self.settings["sample_std"]))
        lomult = 0.9
        himult = 1.1

        for parPrefix, parDefault in parameterNamePrefixAndDefaultsAll.items():
            sampledParameterValues = utils.getSaneNval(
                len(self.withRules),
                lo=lomult * parDefault,
                hi=himult * parDefault,
                mu=parDefault,
                sig=self.settings["sample_std"] * parDefault,
                identicalPars=self.settings["identical_pars"],
            )
            for node, sparval in zip(self.withRules, sampledParameterValues):
                if node in self.genelist:
                    self.par[parPrefix + node] = sparval

        for parPrefix, parDefault in parameterNamePrefixAndDefaultsGenes.items():
            sampledParameterValues = utils.getSaneNval(
                len(self.withRules),
                lo=lomult * parDefault,
                hi=himult * parDefault,
                mu=parDefault,
                sig=self.settings["sample_std"] * parDefault,
                identicalPars=self.settings["identical_pars"],
            )

            for node, sparval in zip(self.withRules, sampledParameterValues):
                if node in self.genelist:
                    self.par[parPrefix + node] = sparval

    def createRegulatoryTerms(self, currgene, combinationOfRegulators, regSpecies):
        """Creates the terms in the activation function which governs
        the transcriptional state of the gene.
        If modeltype is 'hill', each term is composed of products of Hill functions, ranging over
        all possible combinations of regulators.
        Similarly, if modeltype is 'heaviside', each term is simply a
        product of the activities of the combinations of regulators, for all combinations.
        Here, the form of a single term is generated for a single combination of regulators.

        :param currgene: Name of the current gene
        :type currgene: str
        :param combinationOfRegulators: a list of all combinations of regulators of currgene
        :type combinationOfRegulators: list
        """
        strengthSpecified = False
        if not self.interactionStrengthDF.empty:
            if currgene in self.interactionStrengthDF["Gene1"].values:
                strengthSpecified = True
                regulatorsWithStrength = set(
                    self.interactionStrengthDF[
                        self.interactionStrengthDF["Gene1"] == currgene
                    ]["Gene2"].values
                )
        if self.settings["modeltype"] == "hill":
            # Create the hill function terms for each regulator
            hills = []
            for reg in combinationOfRegulators:
                if strengthSpecified and (reg in regulatorsWithStrength):
                    hillThresholdName = "k_" + reg + "_" + currgene
                else:
                    hillThresholdName = "k_" + reg

                if reg in regSpecies:
                    hills.append("(p_" + reg + "/" + hillThresholdName + ")^n_" + reg)
                else:
                    # Note: Only proteins can be regulators
                    hills.append("(" + reg + "/" + hillThresholdName + ")^n_" + reg)
            mult = "*".join(hills)
            return mult
        elif self.settings["modeltype"] == "heaviside":
            terms = []
            for reg in combinationOfRegulators:
                terms.append("p_" + reg)
            mult = "*".join(terms)
            return mult

    def generateModelDict(self):
        """
        Take a DataFrame object with Boolean rules,
        construct ODE equations for each variable.
        This is the core function in BoolODE.
        In a nutshell, a temporary namespace called boolodespace
        is created. In this namespace, the nodes  in the Boolean network
        are first initialized to 0, or OFF. Then, for each gene/node,
        its corresponding rule is evaluated by setting its regulators to
        all combinations binary states. The outcome of each of these Boolean
        rule evaluations is used to decide the value of the activation strength
        parameter 'a' in Hill functions, or the interaction parameter 'w' in
        the Heaviside functions.
        Each step of this conversion is documented in the code directly.
        """

        ##########################################################
        ## Assign values to alpha parameters representing the logic
        ## relationships between variables
        # Initialize new namespace
        boolodespace = {}

        # Initialize empty equations
        for node in self.withRules:
            # Initialize species to 0
            tempStr = node + " = 0"
            exec(tempStr, boolodespace)

        # If there is no rule correspondng to a node, it is either
        # a user specified parameter input, or it is assigned a self loop.
        if not self.parameterInputsDF.empty:
            self.inputs = set(self.withoutRules)
            for k in self.parameterInputsDF["Input"]:
                # Initialize variables to 0
                tempStr = k + " = 0"
                exec(tempStr, boolodespace)

        for i, row in self.df.iterrows():
            # Basal expression:
            # Execute the rule to figure out
            # the value of alpha_0 or omega_0
            exec("booleval = " + row["Rule"], boolodespace)
            if self.settings["modeltype"] == "hill":
                self.par["alpha_" + row["Gene"]] = int(boolodespace["booleval"])
            elif self.settings["modeltype"] == "heaviside":
                self.par["omega_" + row["Gene"]] = utils.heavisideThreshold(
                    boolodespace["booleval"]
                )
        # End initialization

        # Assign values to logic parameters, and construct expressions
        for i, row in self.df.iterrows():
            ## Parse Boolean rule to get list of regulators
            allreg, regSpecies, regInputs = utils.getRegulatorsInRule(
                row["Rule"], self.withRules, self.inputs
            )

            # Basal expression term
            currgene = row["Gene"]
            if self.settings["modeltype"] == "hill":
                num = "( alpha_" + currgene
                den = "( 1"
            elif self.settings["modeltype"] == "heaviside":
                exponent = "- sigmaH_" + currgene + "*( omega_" + currgene

            # Loop over combinations of regulators
            for i in range(1, len(allreg) + 1):
                for combinationOfRegulators in combinations(allreg, i):
                    regulatorExpression = self.createRegulatoryTerms(
                        currgene, combinationOfRegulators, regSpecies
                    )
                    if self.settings["modeltype"] == "hill":
                        # Create Numerator and Denominator
                        den += " +" + regulatorExpression
                        num += (
                            " + a_"
                            + currgene
                            + "_"
                            + "_".join(list(combinationOfRegulators))
                            + "*"
                            + regulatorExpression
                        )
                    elif self.settings["modeltype"] == "heaviside":
                        exponent += (
                            " + w_"
                            + currgene
                            + "_"
                            + "_".join(list(combinationOfRegulators))
                            + "*"
                            + regulatorExpression
                        )

                    # evaluate rule to assign values to parameters
                    ##################################################
                    for node in self.withRules:  #
                        exec(node + " = 0", boolodespace)  #
                    # Set each regulator to ON, evaluate rule. we are looping
                    # over all such combinations of regulators
                    for geneInList in combinationOfRegulators:  #
                        exec(geneInList + " = 1", boolodespace)  #
                        #
                    exec("boolval = " + row["Rule"], boolodespace)  #
                    ##################################################

                    if self.settings["modeltype"] == "hill":
                        self.par[
                            "a_"
                            + currgene
                            + "_"
                            + "_".join(list(combinationOfRegulators))
                        ] = int(boolodespace["boolval"])
                    elif self.settings["modeltype"] == "heaviside":
                        self.par[
                            "w_"
                            + currgene
                            + "_"
                            + "_".join(list(combinationOfRegulators))
                        ] = self.kineticParameterDefaults[
                            "heavisideOmega"
                        ] * utils.heavisideThreshold(
                            boolodespace["boolval"]
                        )

            # Close expressions
            if self.settings["modeltype"] == "hill":
                num += " )"
                den += " )"
                f = "(" + num + "/" + den + ")"
            elif self.settings["modeltype"] == "heaviside":
                # In the case of heaviside expressions, to prevent
                # numerical blowup, we trucate the magnitude of the
                # regulatory terms
                exponent += ")"
                maxexp = "10."  # '100'
                f = (
                    "(1./(1. + np.exp(np.sign("
                    + exponent
                    + ")*min("
                    + maxexp
                    + ",abs("
                    + exponent
                    + ")))))"
                )

            if currgene in self.proteinlist:
                Production = f
                Degradation = "p_" + currgene
                self.varspecs["p_" + currgene] = (
                    "signalingtimescale*(y_max*" + Production + "-" + Degradation + ")"
                )
            else:
                Production = "m_" + currgene + "*" + f
                Degradation = "l_x_" + currgene + "*x_" + currgene
                self.varspecs["x_" + currgene] = Production + "-" + Degradation
                # Create the corresponding translated protein equation
                self.varspecs["p_" + currgene] = (
                    "r_"
                    + currgene
                    + "*"
                    + "x_"
                    + currgene
                    + "- l_p_"
                    + currgene
                    + "*"
                    + "p_"
                    + currgene
                )

        ##########################################################

        # Initialize variables between 0 and 1, Doesn't matter.
        xvals = [1.0 for _ in range(len(self.genelist))]
        pvals = [20.0 for _ in range(len(self.proteinlist))]
        ics = {}

        for node, xv in zip(self.genelist, xvals):
            ics["x_" + node] = xv
            ics["p_" + node] = 0
        for node, pv in zip(self.proteinlist, pvals):
            ics["p_" + node] = pv

        self.ModelSpec["varspecs"] = self.varspecs
        self.ModelSpec["pars"] = self.par
        self.ModelSpec["ics"] = ics

        self.varmapper = {
            i: var for i, var in enumerate(self.ModelSpec["varspecs"].keys())
        }
        self.parmapper = {i: par for i, par in enumerate(self.ModelSpec["pars"].keys())}

    def writeModelToFile(self):
        """
        Writes model to file as a python function.
        The ODE model generated using generateModelDict() is defined as
        a python ODE function called Model(). Model() takes 3 arguments:

        1. The current model state vector Y
        2. The current time t
        3. A list of parameters pars

        The function returns a vector of time derivatives computed from the ODEs.
        Model() is written model.py in the directory of the current job
        """
        self.path_to_ode_model = self.settings["outprefix"] / "model.py"

        with open(self.path_to_ode_model, "w") as out:
            out.write("#####################################################\n")
            out.write("import numpy as np\n")
            out.write("# This file is created automatically\n")
            out.write("def Model(Y,t,pars):\n")
            out.write("    # Parameters\n")
            par_names = sorted(self.ModelSpec["pars"].keys())
            for i, p in enumerate(par_names):
                out.write("    " + p + " = pars[" + str(i) + "]\n")
            outstr = ""
            out.write("    # Variables\n")
            for i in range(len(self.varmapper.keys())):
                out.write("    " + self.varmapper[i] + " = Y[" + str(i) + "]\n")
                outstr += "d" + self.varmapper[i] + ","
            for i in range(len(self.varmapper.keys())):
                vdef = self.ModelSpec["varspecs"][self.varmapper[i]]
                vdef = vdef.replace("^", "**")
                out.write("    d" + self.varmapper[i] + " = " + vdef + "\n")

            out.write("    dY = np.array([" + outstr + "])\n")
            out.write("    return(dY)\n")
            out.write("#####################################################")

    def writeParametersToFile(self):
        """
        Writes dictionary of parameters to file.
        """
        with open(str(self.settings["outprefix"]) + "/parameters.txt", "w") as out:
            out.write("# Automatically generated by BoolODE\n")
            for k, v in self.ModelSpec["pars"].items():
                out.write(k + "\t" + str(v) + "\n")
